/*==============================================================================
   REP_EXP.C
   Report Ease: Expression input and evauation routines.

   ReportEase Plus
   Sub Systems, Inc.
   ReportEase Plus, Copyright (c) 1993, Sub Systems, Inc. All Rights Reserved.
   159 Main Street, #8C, Stoneham,  MA 02180 
   (617) 438-8901.

   Software License Agreement  (1993)              
   ----------------------------------
   This license agreement allows the purchaser the right to modify the 
   source code to incorporate in their application.
   The target application must not be distributed as a standalone report writer
   or a mail merge product.
   Sub Systems, Inc. reserves the right to prosecute anybody found to be 
   making illegal copies of the executable software or compiling the source
   code for the purpose of selling there after.

===============================================================================*/
#include "windows.h"

#if defined (_WIN32)
   #if !defined(WIN32)
     #define WIN32
   #endif
#endif
#if !defined(WIN32) 
  #include "print.h"
#endif

#include "stdio.h"
#include "stdlib.h"
#include "ctype.h"
#include "fcntl.h"
#include "io.h"
#include "sys\types.h"
#include "sys\stat.h"
#include "string.h"
#include "setjmp.h"

#include "rep.h"

#define  PREFIX extern
#include "rep1.h"

/******************************************************************************
    ModifyExp:
    Modify an existing expression.  The new expression is returned through
    the second argument.  On successful execution the function returns the
    variable type of the execution result.  On failure, it returns -1.
******************************************************************************/
ModifyExp(int far *OldExp,int far *NewExp,LPSTR title,int ExpType)
{
    char string[LINE_WIDTH+1],TempString[NAME_WIDTH+1];
    int  CurSym,LastSym,CompExp[NAME_WIDTH+2];
    int  len,CurLen,i,CursPos,CursorSym;
    struct StrField OutField;

    CursorSym=OldExp[0];     // symbol to position the cursor on 
    

    // DECODE_EXP:  decode the old expression into an ASCII string 
    len=0;
    string[0]=0;
    CursPos=0;                                   // cursor position within the string 
    i=0;

    while(OldExp[i]!=END_OF_EXP) {
        CurSym=OldExp[i];
        if (CurSym<MAX_FIELDS) {                 // field symbol 
           CurLen=lstrlen(field[CurSym].name)+1;
           if ((len+CurLen)>LINE_WIDTH) break;
           if (field[CurSym].source==SRC_CONST) lstrcat(string,field[CurSym].name);
           else {                                // convert to lower case
              lstrcpy(TempString,field[CurSym].name);
              lstrlwr(TempString);
              lstrcat(string,TempString);
           }
        }
        else if (CurSym<=LAST_OP) {              // operator symbol 
           CurLen=lstrlen(OpSym[CurSym-FIRST_OP])+1;
           if ((len+CurLen)>LINE_WIDTH) break;
           lstrcpy(TempString,OpSym[CurSym-FIRST_OP]);
           lstrlwr(TempString);
           lstrcat(string,TempString);
        }
        
        if (OldExp[i]==CursorSym) CursPos=len;   // cursor position 

        len=strlen(string);
        i++;
    }

    DlgExpString=(LPSTR)string;                   // string to pass to the dialog box 
    DlgExpType=ExpType;                           // value to pass to the dialog box 

    MODIFY_STRING:
    DlgCursPos=CursPos;                           // value to pass to the dialog box 
    if (CallDialogBox("ExpParam",ExpParam,(DWORD)(LPSTR)title)) {
         // PARSE STRING
         if (!GenerateSymbols(string,NewExp,&CursPos,ExpType)) {   // generate symbols from the given string 
             ReleaseSymbols(NewExp);
             goto MODIFY_STRING;
         }
         ReleaseSymbols(OldExp);                      // delete the symbols for the previous expression 
    }
    else {                                            // retain the old expression 
       i=0;
       while(OldExp[i]!=END_OF_EXP) {
          NewExp[i]=OldExp[i];
          i++;
       }
       NewExp[i]=OldExp[i];
    }

    // COMPILE STRING
    LastSym=CompileExp(NewExp,CompExp);               // compile the expression 
    if (LastSym!=END_OF_EXP) goto MODIFY_STRING;      // error encountered

    // TRIAL RUN
    InitField(&OutField);
    
    LastSym=RunExp(CompExp,&OutField,FALSE);          // run the expression and return the output in the field 

    if (LastSym!=END_OF_EXP) goto MODIFY_STRING;      // error encountered

    if (CompExp[0]!=END_OF_EXP && ExpType==EXP_PARM && OutField.type!=TYPE_LOGICAL) {
       MessageBox(hFrWnd,"This Expression Must Evaluate to a TRUE/FALSE Result, Please Reenter",NULL,MB_OK);
       goto MODIFY_STRING;
    }

    // END

    return OutField.type;
}

/******************************************************************************
    GenerateSymbols:
    Generate symbols from a given string.  The symbols expression is returned
    through the second argument.  If an error is detected, the corresponding
    error position is returned by the third argument, and the function returns
    a FALSE value.
******************************************************************************/
GenerateSymbols(LPSTR string,int far *exp,int far *ErrorPos,int ExpType)
{
    int  i,CurLen,ExpPos,LenUsed;
    int  op,FieldIndex;
    char token[NAME_WIDTH+2],SymType;
    struct StrField CurField;


    ExpPos=0;

    CurLen=LenUsed=0;
    
    NEXT_SYMBOL:
    CurLen=CurLen+LenUsed;
    SymType=NextSymbol(&string[CurLen],token,&LenUsed);

    if (SymType=='E') {           // error encountered 
       exp[ExpPos]=END_OF_EXP;
       (*ErrorPos)=CurLen;
       return FALSE;
    }
   
    if (SymType=='Z') {           // end of the string 
       exp[ExpPos]=END_OF_EXP;
       return TRUE;
    }

    if (SymType=='L') {           // left parenthesis 
       exp[ExpPos]=OP_LPARAN;
       ExpPos++;
       goto NEXT_SYMBOL;
    }

    if (SymType=='R') {           // right parenthesis 
       exp[ExpPos]=OP_RPARAN;
       ExpPos++;
       goto NEXT_SYMBOL;
    }

    if (SymType=='C') {           // Comma (argument separater) 
       exp[ExpPos]=OP_COMMA;
       ExpPos++;
       goto NEXT_SYMBOL;
    }

    if (SymType=='O' || SymType=='G') {  // validate the dot operator or function
       for (op=FIRST_OP;op<=LAST_OP;op++) {
          if (lstrcmp(OpSym[op-FIRST_OP],token)==0 && 
             (!OpFooter[op-FIRST_OP] || ExpType==EXP_FOOTER)) {
             exp[ExpPos]=op;
             ExpPos++;
             goto NEXT_SYMBOL;
          }
       }
       wsprintf(msg,"Invalid Function or Dot Operator: %s",(LPSTR)token);
       MessageBox(hFrWnd,msg,NULL,MB_OK);
       exp[ExpPos]=END_OF_EXP;
       (*ErrorPos)=CurLen;
       return FALSE;
    }

    if (SymType=='P') {           // check for short operators 
       for (op=FIRST_OP;op<=LAST_OP;op++) {
          if (lstrcmp(OpSym[op-FIRST_OP],token)==0) {
             exp[ExpPos]=op;
             ExpPos++;
             goto NEXT_SYMBOL;
          }
       }
       wsprintf(msg,"Invalid Operator: %s",(LPSTR)token);
       MessageBox(hFrWnd,msg,NULL,MB_OK);
       exp[ExpPos]=END_OF_EXP;
       (*ErrorPos)=CurLen;
       return FALSE;
    }

    //** the following token types will be returned via a field **
    InitField(&CurField);
    
    lstrcpy(CurField.name,token); 
    
    if (SymType=='F') {           //  System or an application field 
       if (strstr(token,"SYS->")==token) {
          for (i=0;i<MAX_SYS_FIELDS;i++) {
             if (lstrcmp(&token[5],SysField[i].name)==0) {
                CurField.source=SRC_SYS;                     // source of CurField data 
                CurField.type=SysField[i].type;              // num/alpha etc 
                CurField.SysIdx=i;                           // index into the system CurField structure 
                break;
             }
          }
          if (i==MAX_SYS_FIELDS) {
             wsprintf(msg,"Invalid System Field: %s",(LPSTR)token);
             MessageBox(hFrWnd,msg,NULL,MB_OK);
             exp[ExpPos]=END_OF_EXP;
             (*ErrorPos)=CurLen;
             return FALSE;
          }
       }
       else if (strstr(token,"DLG->")==token) {              // dialog field 
          for (i=0;i<MAX_DLGS;i++) {
             if (DlgField[i].InUse && lstrcmp(&token[5],DlgField[i].name)==0) {
                CurField.source=SRC_DLG;                     // source of CurField data 
                CurField.type=DlgField[i].type;              // num/alpha etc 
                CurField.SysIdx=i;                           // index into the system CurField structure 
                break;
             }
          }
          if (i==MAX_DLGS) {
             wsprintf(msg,"Invalid Dialog Field: %s",(LPSTR)token);
             MessageBox(NULL,msg,NULL,MB_OK);
             exp[ExpPos]=END_OF_EXP;
             (*ErrorPos)=CurLen;
             return FALSE;
          }
       }
       else {
          CurField.source=SRC_APPL;  // validate as an application field 
          if (!(FormArg.VerifyField)(&CurField,0)) {  // call a user routine to verify a data field 
             wsprintf(msg,"Invalid Data Field: %s",(LPSTR)token);
             MessageBox(NULL,msg,NULL,MB_OK);
             exp[ExpPos]=END_OF_EXP;
             (*ErrorPos)=CurLen;
             return FALSE;
          }
          if (CurField.type==TYPE_PICT) {       // picture field not allowed in the expression
             wsprintf(msg,"Picture Field Not Allowed: %s",(LPSTR)token);
             MessageBox(NULL,msg,NULL,MB_OK);
             exp[ExpPos]=END_OF_EXP;
             (*ErrorPos)=CurLen;
             return FALSE;
          }
       }
    }

    if (SymType=='S') {           // string constant or date 
       CurField.source=SRC_CONST;
       if (ValidateDate(token,&(CurField.NumData))) {
          if (CurField.NumData<0) {
             wsprintf(msg,"Invalid Date: %s",(LPSTR)token);
             MessageBox(NULL,msg,NULL,MB_OK);
             exp[ExpPos]=END_OF_EXP;
             (*ErrorPos)=CurLen;
             return FALSE;
          }
          else CurField.type=TYPE_DATE;
       
       }
       else {
          CurField.type=TYPE_TEXT;
          CurField.width=strlen(token)+1;
       }
    }

    if (SymType=='N') {           // numeric constant 
       CurField.source=SRC_CONST;
       CurField.type=TYPE_NUM;
       CurField.NumData=atol(token);
    }

    if (SymType=='D') {           // numeric constant 
       CurField.source=SRC_CONST;
       CurField.type=TYPE_DBL;
       CurField.DblData=atof(token);
    }

    // allocate a field slot
    if (END_OF_EXP==(FieldIndex=FindFieldSlot())) {
       MessageBox(hFrWnd,"Ran out of field table!",NULL,MB_OK);
       exp[ExpPos]=END_OF_EXP;
       (*ErrorPos)=CurLen;
       return FALSE;
    }

    // copy the field into the current array position 
    field[FieldIndex]=CurField;
    
    FrModified=TRUE;
    exp[ExpPos]=FieldIndex;
    ExpPos++;
    goto NEXT_SYMBOL;

}

/******************************************************************************
    NextSymbol:
    Extracts the first symbol from a string.  The second argument returns
    the extracted symbol.  The third argument returns the length utilized.
    The function returns one of these codes:
         C = Comma
         S = String Constant
         N = Numeric Constant
         D = Double Constant
         O = Dot Operator
         P = Short operator
         G = Function name
         L = Left paranthesis
         R = Right paranthesis
         F = Field name
         E = Error encountered.
         Z = End of string
******************************************************************************/
char NextSymbol(LPSTR string,LPSTR symbol,int far *LenUsed)
{
    int i,j,k,len,OpLen;

    symbol[0]=0;
    i=0;
    len=strlen(string);
    if (len==0) return 'Z';               // end of string 

    while(i<len && string[i]==' ') i++;   // ignore spaces 
    if (i==len) return 'Z';               // end of string 

    j=0;
    if (string[i]=='\"') {                // extract the string leteral 
       symbol[j]=string[i];
       i++;j++;
       while(i<len && string[i]!='\"') {
          symbol[j]=string[i];
          i++;j++;
       }
       if (i==len) {
          MessageBox(hFrWnd,"Missing Closing Quotes",NULL,MB_OK);
          return 'E';
       }    
       symbol[j]=string[i];
       i++;j++;
       symbol[j]=0;
       if (j>NAME_WIDTH) {
          MessageBox(NULL,"String Too Long!",NULL,MB_OK);
          return 'E';
       }
       (*LenUsed)=i;
       return 'S';                         // string constant 
    }
    else if (string[i]==',') {             // comma (argument separater) 
       symbol[j]=string[i];
       i++;j++;
       symbol[j]=0;
       (*LenUsed)=i;
       return 'C';
    }
    else if (string[i]=='.') {
       if ((i+1)<len && string[i+1]>='0' && string[i+1]<='9') { // double constant 
           symbol[j]=string[i];
           i++;j++;
           while (i<len && string[i]>='0' && string[i]<='9') {
              symbol[j]=string[i];
              i++;j++;
           }
           symbol[j]=0;
           (*LenUsed)=i;
           return 'D';                      // double numeric constant   
       }
       else {                              // check for a dot operator 
           symbol[j]=string[i];
           i++;j++;
           while(i<len && string[i]!='.') {
              symbol[j]=string[i];
              i++;j++;
           }
           if (i==len) {
              MessageBox(hFrWnd,"Missing Dot in the Dot Operator",NULL,MB_OK);
              return 'E';
           }    
           symbol[j]=string[i];
           i++;j++;
           symbol[j]=0;
           lstrupr(symbol);
           (*LenUsed)=i;
           return 'O';                          // dot operator
       }
    }
    else if (string[i]>='0' && string[i]<='9') {// numeric constant 
       symbol[j]=string[i];
       i++;j++;
       while (i<len && string[i]>='0' && string[i]<='9') {
          symbol[j]=string[i];
          i++;j++;
       }
       if (i==len || string[i]!='.') {
          symbol[j]=0;
          (*LenUsed)=i;
          return 'N';                           // numeric constant   
       }
       if ((i+1)<len && string[i+1]>='0' && string[i+1]<='9') { // double constant 
           symbol[j]=string[i];
           i++;j++;
           while (i<len && string[i]>='0' && string[i]<='9') {
              symbol[j]=string[i];
              i++;j++;
           }
           symbol[j]=0;
           (*LenUsed)=i;
           return 'D';                          // double numeric constant   
       }
       symbol[j]=0;
       (*LenUsed)=i;
       return 'N';                              // numeric constant   
    }
    else if (string[i]=='(') {                  // left paranthesis 
       symbol[j]=string[i];
       i++;j++;
       symbol[j]=0;
       (*LenUsed)=i;
       return 'L';
    }
    else if (string[i]==')') {                  // right paranthesis 
       symbol[j]=string[i];
       i++;j++;
       symbol[j]=0;
       (*LenUsed)=i;
       return 'R';
    }
    else {                                      // other symbols 
       OpLen=0;                                 // find the longest operator that fits 
       for (k=0;k<MAX_OPS;k++) {                // check for short operators 
          if (strstr(&string[i],OpSym[k])==&string[i] && !OpFct[k]) {
             if ((int)strlen(OpSym[k])>OpLen) {
                lstrcpy(symbol,OpSym[k]);
                OpLen=strlen(symbol);
             }
          } 
       }
       if (OpLen>0) {                           // atleast one operator found 
          i=i+OpLen;
          (*LenUsed)=i;
          return 'P';                           // short operators 
       }

       symbol[j]=string[i];
       i++;j++;
       while(i<len) {
          if (strchr("()*/+#<=\"\' .$,",string[i])) break;     // invalid field characters 
          if (string[i]=='-' && string[i+1] != '>') break;
          if (i>0 && string[i]=='>' && string[i-1] != '-') break;

          symbol[j]=string[i];
          i++;j++;
       }
       symbol[j]=0;
       lstrupr(symbol);
       (*LenUsed)=i;
       
       if (i<len && string[i]=='(') return 'G'; // function name 
       else                         return 'F'; // field 
    }
}

/******************************************************************************
    CompileExp:
    Compile the given expression. 
    The function returns the last symbol processed from the input
    expression.  
******************************************************************************/
int CompileExp(int far *InExp,int far *OutExp)
{
    int stack[NAME_WIDTH+2],StackSym,InSym,TempExp[NAME_WIDTH+2];
    int i,StackPtr,InIdx,OutIdx,OpIdx,InPrec,StackPrec,FieldCount,
        OpCount,AfterLeftParan;


    //****************** convert function into operators ******************
    InIdx=0;
    OutIdx=0;
    StackPtr=0;
    while(InExp[InIdx]!=END_OF_EXP) {
        InSym=InExp[InIdx];
        
        if (InSym>=FIRST_OP && InSym<=LAST_OP) {
           OpIdx=InSym-FIRST_OP;                  // index into the operator table 
           if (OpFct[OpIdx] && !OpUnary[OpIdx]) { // two argument function symbol 
              if (!PushStack(stack,InSym,&StackPtr)) return InSym;  // push the function name 
           }
           else if (InSym==OP_COMMA) {            // substitute by the functio name 
              if (!PopStack(stack,&StackSym,&StackPtr)) return InSym;
              TempExp[OutIdx]=StackSym;
              OutIdx++;
           }
           else {
              TempExp[OutIdx]=InSym;
              OutIdx++;
           }
        }
        else {
           TempExp[OutIdx]=InSym;
           OutIdx++;
        }
        InIdx++;
    }
    TempExp[OutIdx]=END_OF_EXP;                    // last character 

    if (StackPtr!=0) {                            // stack symbol not used (error) 
       PopStack(stack,&StackSym,&StackPtr);
       MessageBox(hFrWnd,"Invalid Expression!",NULL,MB_OK);
       return StackSym;
    }


    //*********************** convert to postfix string ********************
    InIdx=0;
    OutIdx=0;
    StackPtr=0;
    InSym=TempExp[0];

    if (!PushStack(stack,OP_LPARAN,&StackPtr)) return InSym;  // push a left paran 
    AfterLeftParan=TRUE;

    while(TempExp[InIdx]!=END_OF_EXP) {
        InSym=TempExp[InIdx];

        // process fields 
        if (InSym<MAX_FIELDS) {            // transfer to output 
           OutExp[OutIdx]=InSym;
           OutIdx++;
           goto NEXT_SYM;
        }

        OpIdx=TempExp[InIdx]-FIRST_OP;     // index into the operator table 
        
        // output a dummy field for unary operators 
        if (OpUnary[OpIdx]) {
           OutExp[OutIdx]=FLD_DUMMY;
           OutIdx++;
        }

        // special treatment of + and - as unary operators 
        if (AfterLeftParan && (InSym==OP_PLUS || InSym==OP_MINUS)) {
           OutExp[OutIdx]=FLD_DUMMY;
           OutIdx++;
        }
        
        // process parenthesis operators 
        
        if (InSym==OP_LPARAN) {
           if (!PushStack(stack,InSym,&StackPtr)) return InSym; // push a left paran 
           AfterLeftParan=TRUE;
           InIdx++;
           continue;
        }

        if (InSym==OP_RPARAN) {            // transfer operators from stack to output 
           while (TRUE) {
              if (!PopStack(stack,&StackSym,&StackPtr)) return InSym;
              if (StackSym==OP_LPARAN) break;
              OutExp[OutIdx]=StackSym;
              OutIdx++;
           }
           goto NEXT_SYM;
        }

        // process other operators 
        InPrec=OpPrec[OpIdx];              // precedence of the current operator 
        
        while (TRUE) {                     // pop of equal or higher precedence operators 
           if (!PopStack(stack,&StackSym,&StackPtr)) return InSym;
           
           StackPrec=OpPrec[StackSym-FIRST_OP];  // precedence of the operator on the stack 
           if (StackPrec<InPrec || StackSym==OP_LPARAN) {  // push it back 
              if (!PushStack(stack,StackSym,&StackPtr)) return InSym;
              break;
           }
           OutExp[OutIdx]=StackSym;        // transfer to output 
           OutIdx++;
        }
        
        // push the current operator to the stack 
        if (!PushStack(stack,InSym,&StackPtr)) return InSym;
        
        NEXT_SYM:
        InIdx++;
        AfterLeftParan=FALSE;
    }
    
    // transfer remaining stack symbols to output 
    while (TRUE) {                     // pop of equal or higher precedence operators 
       if (!PopStack(stack,&StackSym,&StackPtr)) return InSym;
       
       if (StackSym==OP_LPARAN) {
          if (StackPtr==0) break;
          else {
             MessageBox(hFrWnd,"Parenthesis Mismatch!",NULL,MB_OK);
             return InSym;
          }
       }
       
       OutExp[OutIdx]=StackSym;        // transfer to output 
       OutIdx++;
    }
    

    OutExp[OutIdx]=END_OF_EXP;         // terminate the compiled string 
    
    if (OutIdx==0) return END_OF_EXP;

    
    // VERIFY EXPPRESSION
    FieldCount=OpCount=i=0;
    
    while(OutExp[i]!=END_OF_EXP) {
        if (OutExp[i]<MAX_FIELDS || OutExp[i]==FLD_DUMMY) FieldCount++;
        else                                              OpCount++;
        i++;
    }
    FieldCount--;                      // number of operators should be one less than number of fields 

    if (FieldCount<OpCount) {
        MessageBox(hFrWnd,"A Field Missing for an Operator",NULL,MB_OK);
        return InSym;
    }
    if (FieldCount>OpCount) {
        MessageBox(hFrWnd,"An Operator Missing for a Field",NULL,MB_OK);
        return InSym;
    }

    return END_OF_EXP;
}

/******************************************************************************
    PushStack:
    Push a symbol on the stack.
******************************************************************************/
PushStack(int far *stack,int sym,int far *StackPtr)
{
    if ((*StackPtr)>=NAME_WIDTH) {
       MessageBox(hFrWnd,"ReportEase Stack Overflow!",NULL,MB_OK);
       return FALSE;
    } 
    stack[(*StackPtr)]=sym;
    (*StackPtr)++;
    return TRUE;
}

/******************************************************************************
    PopStack:
    Pop a symbol from the stack.
******************************************************************************/
PopStack(int far *stack,int far *sym,int far *StackPtr)
{
    if ((*StackPtr)<=0) {
       MessageBox(hFrWnd,"Parenthesis Mismatch or Stack Underflow!",NULL,MB_OK);
       return FALSE;
    } 
    (*StackPtr)--;
    (*sym)=stack[(*StackPtr)];
    return TRUE;
}

/******************************************************************************
    RunExp:
    This routines executes a compiled expression (exp). The calculated
    value is returned in the 'OutField' parameter.  The last parameter if
    FALSE instructs the routine to do dummy calculations without using the
    actual values in the fields.
    The funtion returns END_OF_EXP on a successful execution, otherwise
    it returns the value of the symbol where the error occurred.
******************************************************************************/
int RunExp(int far *OrigExp,struct StrField huge *OutField,int calc)
{
    struct StrField AnsField,huge *ResField;
    int i,OpIdx,TotalOperands,len;
    int exp[NAME_WIDTH+2],OperandIdx[2]={0,0},CurSym,operand[2],TempIdx;

    if (OrigExp[0]==END_OF_EXP) return END_OF_EXP;
    
    for (i=0;i<MAX_TEMP_FIELDS;i++) {
       TempField[i].InUse=FALSE;
       TempField[i].CharData=NULL;
    }
    AnsField.InUse=TRUE;
    AnsField.CharData=NULL;

    // make a working copy of the expression 
    i=0;
    while(OrigExp[i]!=END_OF_EXP) {
        exp[i]=OrigExp[i];
        i++;
    }
    exp[i]=OrigExp[i];

    // find the next operator and operands
    NEXT_CALC:
    TotalOperands=0;
    for (OpIdx=0;exp[OpIdx]!=END_OF_EXP;OpIdx++) {
        CurSym=exp[OpIdx];
        if (CurSym==SYM_NOP) continue;   // no operation symbol 
        if (CurSym>=FIRST_OP && CurSym<=LAST_OP) break;
        
        if (CurSym<MAX_FIELDS || CurSym==FLD_DUMMY || (CurSym>=FIRST_TEMP_FIELD && CurSym<=LAST_TEMP_FIELD)) {
           if (TotalOperands==2) {       // scroll the operand table 
              OperandIdx[0]=OperandIdx[1];
              TotalOperands=1;
           }
           OperandIdx[TotalOperands]=OpIdx;
           TotalOperands++;
        }
    }

    // isolate operands 
    for (i=0;i<TotalOperands;i++) operand[i]=exp[OperandIdx[i]];
   
    // Run complete, pass the result 
    if (exp[OpIdx]==END_OF_EXP) {        // no operator found 
        if (TotalOperands==1) {          // first operand is the result 
           if (operand[0]>=FIRST_TEMP_FIELD) ResField=&(TempField[operand[0]-FIRST_TEMP_FIELD]);
           else                              ResField=&(field[operand[0]]);
           
           OutField->type=ResField->type;
           OutField->NumData=ResField->NumData;
           OutField->DblData=ResField->DblData;

           if (calc) {
              if (ResField->type==TYPE_TEXT) {
                 if (ResField->CharData==NULL) len=0;
                 else                          len=lstrlen(ResField->CharData);
                 if (NULL==(OutField->CharData=OurAlloc(len+2))) {
                    MessageBox(hFrWnd,"Ran out of Memory During Expression Calculation",NULL,MB_OK);
                    return operand[0];
                 }
                 if (ResField->CharData==NULL) OutField->CharData[0]=0;
                 else lstrcpy(OutField->CharData,ResField->CharData);
              }
           }
           else OutField->CharData=NULL;
           
           if (operand[0]>=FIRST_TEMP_FIELD) {   // freeup the memory used by temporary field 
              if (ResField->type==TYPE_TEXT && ResField->CharData!=NULL) {
                 OurFree(ResField->CharData);
                 ResField->CharData=NULL;
              }
           }

           return END_OF_EXP;
        }
        MessageBox(hFrWnd,"Too Many or Too Few Fields",NULL,MB_OK);
        return operand[0];
    }

    if (TotalOperands<2) {
        MessageBox(hFrWnd,"Invalid Operator/Operand Position",NULL,MB_OK);
        return exp[OpIdx];
    }


    if (operand[1]==FLD_DUMMY) {
        MessageBox(hFrWnd,"Invalid Unary Operator Position",NULL,MB_OK);
        return exp[OpIdx];
    }
    
    // calculate 

    if (!calculate(operand[0],operand[1],exp[OpIdx],&AnsField,calc)) return exp[OpIdx];

    // locate a temporary field for storage 
    for (TempIdx=0;TempIdx<MAX_TEMP_FIELDS;TempIdx++) {
       if (!TempField[TempIdx].InUse) break;
    }
    
    if (TempIdx==MAX_TEMP_FIELDS) {
       MessageBox(hFrWnd,"Ran Out of Temporary Field Buffers",NULL,MB_OK);
       return exp[OpIdx];
    }
       
    TempField[TempIdx]=AnsField;
    AnsField.CharData=NULL;           // pointer transferred, if any 
    TempField[TempIdx].InUse=TRUE;
    
    exp[OperandIdx[0]]=SYM_NOP;       // mark these fields as used up 
    exp[OperandIdx[1]]=SYM_NOP;

    exp[OpIdx]=TempIdx+FIRST_TEMP_FIELD; // holds the result of this calculation 

    goto NEXT_CALC;

}

/******************************************************************************
    calculate:
    This routine applies the given operator on two given fields and
    returns the result in a field structure.
******************************************************************************/
calculate(int fld1,int fld2, int op, struct StrField huge *AnsField, int calc)
{
    struct StrField huge *f1,huge *f2,DummyField;
    int  type,type1,type2,result;

    // initialize the answer field
    if (calc) InitField(AnsField);

    // build the pointers to the first and second fields 
    if (fld2==FLD_DUMMY) {
       MessageBox(hFrWnd,"Field Type Mismatch!",NULL,MB_OK);
       return FALSE;
    }
    else {
       if (fld2<MAX_FIELDS)   f2=&field[fld2];
       else {
          f2=&TempField[fld2-FIRST_TEMP_FIELD];
          f2->InUse=FALSE;    // make it available for next time 
       }
    }
    
    if (fld1==FLD_DUMMY) {
       f1=(struct StrField huge *)&DummyField;
       InitField(f1);
       f1->type=f2->type;
    }    
    else {
       if (fld1<MAX_FIELDS)   f1=&field[fld1];
       else {
          f1=&TempField[fld1-FIRST_TEMP_FIELD];
          f1->InUse=FALSE;    // make it available for next time 
       }
    }
    
    type1=f1->type;
    type2=f2->type;

    if (OpFct[op-FIRST_OP] || op==OP_THEN || op==OP_ELSE) goto SKIP_TYPE_CHECK;

    // ensure that the argument types match 
    if (type1==TYPE_DBL) type1=TYPE_NUM;
    if (type2==TYPE_DBL) type2=TYPE_NUM;
    if (type1!=type2) {
       MessageBox(hFrWnd,"Field Type Mismatch!",NULL,MB_OK);
       return FALSE;
    }
    type=type1;            // common argument type 

    if (type==TYPE_NUM && f1->type==TYPE_DBL && f2->type!=TYPE_DBL) {
       f2->DblData=(double)f2->NumData;
    }
    if (type==TYPE_NUM && f2->type==TYPE_DBL && f1->type!=TYPE_DBL) {
       f1->DblData=(double)f1->NumData;
    }

    if (f1->type==TYPE_DBL || f2->type==TYPE_DBL) AnsField->type=TYPE_DBL;
    else                                          AnsField->type=type;

    SKIP_TYPE_CHECK:

    if (f1->type==TYPE_DATE && f1->NumData<=999999L) f1->NumData=19000000L+f1->NumData;
    if (f2->type==TYPE_DATE && f2->NumData<=999999L) f2->NumData=19000000L+f2->NumData;


    //************************ perform calculations ************************
    result=calculate1(f1,f2,AnsField,op,type,calc);                             // perform logical operations 
    if (result==ERR_INCOMPLETE) result=calculate2(f1,f2,AnsField,op,type,calc); // perform other than logical operations 
    if (result==ERR_INCOMPLETE) result=calculate3(f1,f2,AnsField,op,type1,type2,calc); // perform other than logical operations 

    if (result==ERR_NO_MEM) { 
       MessageBox(hFrWnd,"Ran Out of Memory for String Calculation!",NULL,MB_OK);
       return FALSE;
    }
    else if (result==ERR_ZERO_DIVIDE) {
       MessageBox(hFrWnd,"Error: Division by Zero!",NULL,MB_OK);
       return FALSE;
    
    }
    else if (result==ERR_TYPE_MISMATCH) {
       MessageBox(hFrWnd,"Field Type Mismatch!",NULL,MB_OK);
       return FALSE;
    }


    // END
    // release character string space for the temporary fields 
    if (fld1>=FIRST_TEMP_FIELD && fld1<=LAST_TEMP_FIELD) {
       if (calc && f1->type==TYPE_TEXT && f1->CharData!=NULL) {
          OurFree(f1->CharData);
          f1->CharData=NULL;
       }
    }

    if (fld2>=FIRST_TEMP_FIELD && fld2<=LAST_TEMP_FIELD) {
       if (calc && f2->type==TYPE_TEXT && f2->CharData!=NULL) {
          OurFree(f2->CharData);
          f2->CharData=NULL;
       }
    }
    
    
    return TRUE;
}

/******************************************************************************
    calculate1:
    This routine performs calculation for a logical operator.
    The first and second argument specify the two input field. The third
    argument is used to return the result.
    This function returns 0 if the operator is processed, otherwise
    it returns an error code.
******************************************************************************/
calculate1(struct StrField huge *f1,struct StrField huge *f2,struct StrField huge *AnsField,int op,int type,int calc)
{

    switch (op) {
       case  OP_OR:
             if (type==TYPE_LOGICAL) {
                if (f1->NumData || f2->NumData) AnsField->NumData=TRUE;
                else                            AnsField->NumData=FALSE;
             }       
             else  return ERR_TYPE_MISMATCH;
             break;

       case  OP_AND:
             if (type==TYPE_LOGICAL) {
                if (f1->NumData && f2->NumData) AnsField->NumData=TRUE;
                else                            AnsField->NumData=FALSE;
             }       
             else  return ERR_TYPE_MISMATCH;
             break;

       case  OP_EQU:
             AnsField->type=TYPE_LOGICAL;
             AnsField->NumData=FALSE;
             if (type==TYPE_TEXT) {
                if (calc) {
                   if (f1->CharData!=NULL && f2->CharData!=NULL) {
                       rTrim(f1->CharData);
                       rTrim(f2->CharData);
                       if (lstrcmp(f1->CharData,f2->CharData)==0) AnsField->NumData=TRUE;
                   }
                }
             }
             else if (type==TYPE_NUM) {
                if (f1->type==TYPE_NUM && f2->type==TYPE_NUM) {
                   if (f1->NumData==f2->NumData) AnsField->NumData=TRUE;
                }
                else if (f1->DblData==f2->DblData) AnsField->NumData=TRUE;
             }
             else if (type==TYPE_LOGICAL || type==TYPE_DATE) {
                if (f1->NumData==f2->NumData) AnsField->NumData=TRUE;
             }
             else  return ERR_TYPE_MISMATCH;
             break;

       case  OP_NOT_EQU:
       case  OP_NOT_EQU2:
             AnsField->type=TYPE_LOGICAL;
             AnsField->NumData=TRUE;
             if (type==TYPE_TEXT) {
                if (calc) {
                   if (f1->CharData!=NULL && f2->CharData!=NULL) {
                       rTrim(f1->CharData);
                       rTrim(f2->CharData);
                       if (lstrcmp(f1->CharData,f2->CharData)==0) AnsField->NumData=FALSE;
                   }
                }
             }
             else if (type==TYPE_NUM) {
                if (f1->type==TYPE_NUM && f2->type==TYPE_NUM) {
                   if (f1->NumData==f2->NumData) AnsField->NumData=FALSE;
                }
                else if (f1->DblData==f2->DblData) AnsField->NumData=FALSE;
             }
             else if (type==TYPE_LOGICAL || type==TYPE_DATE) {
                if (f1->NumData==f2->NumData) AnsField->NumData=FALSE;
             }
             else  return ERR_TYPE_MISMATCH;
             break;

       case  OP_GT:
             AnsField->type=TYPE_LOGICAL;
             AnsField->NumData=FALSE;
             if (type==TYPE_TEXT) {
                if (calc) {
                   if (f1->CharData!=NULL && f2->CharData!=NULL) {
                       rTrim(f1->CharData);
                       rTrim(f2->CharData);
                       if (lstrcmp(f1->CharData,f2->CharData)>0) AnsField->NumData=TRUE;
                   }
                }
             }
             else if (type==TYPE_NUM) {
                if (f1->type==TYPE_NUM && f2->type==TYPE_NUM) {
                   if (f1->NumData>f2->NumData) AnsField->NumData=TRUE;
                }
                else if (f1->DblData>f2->DblData) AnsField->NumData=TRUE;
             }
             else if (type==TYPE_DATE) {
                if (f1->NumData>f2->NumData) AnsField->NumData=TRUE;
             }
             else  return ERR_TYPE_MISMATCH;
             break;

       case  OP_LT:
             AnsField->type=TYPE_LOGICAL;
             AnsField->NumData=FALSE;
             if (type==TYPE_TEXT) {
                if (calc) {
                   if (f1->CharData!=NULL && f2->CharData!=NULL) {
                       rTrim(f1->CharData);
                       rTrim(f2->CharData);
                       if (lstrcmp(f1->CharData,f2->CharData)<0) AnsField->NumData=TRUE;
                   }
                }
             }
             else if (type==TYPE_NUM) {
                if (f1->type==TYPE_NUM && f2->type==TYPE_NUM) {
                   if (f1->NumData<f2->NumData) AnsField->NumData=TRUE;
                }
                else if (f1->DblData<f2->DblData) AnsField->NumData=TRUE;
             }
             else if (type==TYPE_DATE) {
                if (f1->NumData<f2->NumData) AnsField->NumData=TRUE;
             }
             else  return ERR_TYPE_MISMATCH;
             break;

       case  OP_GTE:
             AnsField->type=TYPE_LOGICAL;
             AnsField->NumData=FALSE;
             if (type==TYPE_TEXT) {
                if (calc) {
                   if (f1->CharData!=NULL && f2->CharData!=NULL) {
                       rTrim(f1->CharData);
                       rTrim(f2->CharData);
                       if (lstrcmp(f1->CharData,f2->CharData)>=0) AnsField->NumData=TRUE;
                   }
                }
             }
             else if (type==TYPE_NUM) {
                if (f1->type==TYPE_NUM && f2->type==TYPE_NUM) {
                   if (f1->NumData>=f2->NumData) AnsField->NumData=TRUE;
                }
                else if (f1->DblData>=f2->DblData) AnsField->NumData=TRUE;
             }
             else if (type==TYPE_DATE) {
                if (f1->NumData>=f2->NumData) AnsField->NumData=TRUE;
             }
             else  return ERR_TYPE_MISMATCH;
             break;

       case  OP_LTE:
             AnsField->type=TYPE_LOGICAL;
             AnsField->NumData=FALSE;
             if (type==TYPE_TEXT) {
                if (calc) {
                   if (f1->CharData!=NULL && f2->CharData!=NULL) {
                       rTrim(f1->CharData);
                       rTrim(f2->CharData);
                       if (lstrcmp(f1->CharData,f2->CharData)<=0) AnsField->NumData=TRUE;
                   }
                }
             }
             else if (type==TYPE_NUM) {
                if (f1->type==TYPE_NUM && f2->type==TYPE_NUM) {
                   if (f1->NumData<=f2->NumData) AnsField->NumData=TRUE;
                }
                else if (f1->DblData<=f2->DblData) AnsField->NumData=TRUE;
             }
             else if (type==TYPE_DATE) {
                if (f1->NumData<=f2->NumData) AnsField->NumData=TRUE;
             }
             else  return ERR_TYPE_MISMATCH;
             break;

       case  OP_PART:
             if (type==TYPE_TEXT) {
                AnsField->type=TYPE_LOGICAL;
                AnsField->NumData=FALSE;
                if (calc) {
                   if (f1->CharData!=NULL && f2->CharData!=NULL) {
                      if (FarStringIndex(f2->CharData,f1->CharData)==(UINT)lstrlen(f2->CharData)) 
                            AnsField->NumData=FALSE;
                      else  AnsField->NumData=TRUE;
                   }
                }
             }
             else  return ERR_TYPE_MISMATCH;
             break;
       default:
             return ERR_INCOMPLETE;         // operator not processed 
    }

     return 0;
}

/******************************************************************************
    calculate2:
    This routine performs calculation for all operators except logical 
    operators.
    The first and second argument specify the two input field. The third
    argument is used to return the result.
    This function returns 0 if the operator is processed, otherwise
    it returns an error code.
******************************************************************************/
calculate2(struct StrField huge *f1,struct StrField huge *f2,struct StrField huge *AnsField,int op,int type,int calc)
{
    struct StrField huge *fld;
    int  len,AdjLen,i;
       
    switch (op) {
       case  OP_PLUS:
             if (type==TYPE_TEXT) {
                if (calc) {
                   if (f1->CharData!=NULL && f2->CharData!=NULL) {
                      len=lstrlen(f1->CharData)+lstrlen(f2->CharData);
                      if (NULL==(AnsField->CharData=OurAlloc(len+2))) return ERR_NO_MEM;
                      lstrcpy(AnsField->CharData,f1->CharData);
                      lstrcat(AnsField->CharData,f2->CharData);
                   }
                }
             }
             else if (type==TYPE_NUM) {
                AnsField->NumData=f1->NumData;
                AnsField->DblData=f1->DblData;
                (AnsField->NumData)+=f2->NumData;
                (AnsField->DblData)+=f2->DblData;
             }       
             else  return ERR_TYPE_MISMATCH;

             break;

       case  OP_MINUS:
             if (type==TYPE_TEXT) {
                if (calc) {
                   if (f1->CharData!=NULL && f2->CharData!=NULL) {
                      len=lstrlen(f1->CharData)+lstrlen(f2->CharData);
                      if (NULL==(AnsField->CharData=OurAlloc(len+2))) return ERR_NO_MEM;
                      lstrcpy(AnsField->CharData,f1->CharData);
                      AdjLen=lstrlen(AnsField->CharData);
                      rTrim(AnsField->CharData);
                      AdjLen=AdjLen-lstrlen(AnsField->CharData);
                      lstrcat(AnsField->CharData,f2->CharData);
                      if (AdjLen>0) {            // add to the right of the string 
                         len=lstrlen(AnsField->CharData);
                         for (i=0;i<AdjLen;i++) AnsField->CharData[len+i]=' ';
                         AnsField->CharData[len+AdjLen]=0;
                      }
                   }
                }
             }
             else if (type==TYPE_NUM) {
                AnsField->NumData=f1->NumData;
                AnsField->DblData=f1->DblData;
                (AnsField->NumData)-=f2->NumData;
                (AnsField->DblData)-=f2->DblData;
             }       
             else  return ERR_TYPE_MISMATCH;

             break;

       case  OP_MULT:
             if (type==TYPE_NUM) {
                AnsField->NumData=f1->NumData;
                AnsField->DblData=f1->DblData;
                (AnsField->NumData)*=f2->NumData;
                (AnsField->DblData)*=f2->DblData;
             }       
             else  return ERR_TYPE_MISMATCH;
             break;

       case  OP_DIV:
             if (type==TYPE_NUM) {
                   AnsField->NumData=f1->NumData;
                   AnsField->DblData=f1->DblData;
                if (calc) {
                   if (f1->type==TYPE_NUM && f2->type==TYPE_NUM) {
                      if (f2->NumData>0) (AnsField->NumData)/=f2->NumData;else return ERR_ZERO_DIVIDE;
                   }
                   else if (f2->DblData>0) (AnsField->DblData)/=f2->DblData;else return ERR_ZERO_DIVIDE;
                }
             }       
             else  return ERR_TYPE_MISMATCH;
             break;

       case  OP_NOT:
             if (type==TYPE_LOGICAL) {
                if (f2->NumData) AnsField->NumData=FALSE;
                else             AnsField->NumData=TRUE;
             }       
             else  return ERR_TYPE_MISMATCH;
             break;

       case  OP_TOT_OF:
             if (type==TYPE_NUM) {
                f2->SumType=SUM_TOTAL;     // accumulate totals for this field 
                AnsField->NumData=f2->HoldNum;
                AnsField->DblData=f2->HoldDbl;
             }       
             else  return ERR_TYPE_MISMATCH;
             break;

       case  OP_AVE_OF:
             if (type==TYPE_NUM) {
                f2->SumType=SUM_AVERAGE;   // accumulate totals and count for this field 
                AnsField->NumData=0;
                AnsField->DblData=0;
                if (f2->count>0) AnsField->NumData=f2->HoldNum/f2->count;
                if (f2->count>0) AnsField->DblData=f2->HoldDbl/f2->count;
             }       
             else  return ERR_TYPE_MISMATCH;
             break;

       case  OP_MAX_OF:
             if (type==TYPE_NUM) {
                f2->SumType=SUM_MAX;      // accumulate max for this field 
                AnsField->NumData=f2->HoldNum;
                AnsField->DblData=f2->HoldDbl;
             }       
             else  return ERR_TYPE_MISMATCH;
             break;

       case  OP_MIN_OF:
             if (type==TYPE_NUM) {
                f2->SumType=SUM_MIN;      // accumulate min for this field 
                AnsField->NumData=f2->HoldNum;
                AnsField->DblData=f2->HoldDbl;
             }       
             else  return ERR_TYPE_MISMATCH;
             break;

       case  OP_COUNT_OF:
             if (type==TYPE_NUM) {
                f2->SumType=SUM_COUNT;      // accumulate count for this field 
                AnsField->type=TYPE_NUM;
                AnsField->NumData=f2->count;
             }       
             else  return ERR_TYPE_MISMATCH;
             break;

       case  OP_IF:
             if (f2->type==TYPE_LOGICAL) {
                AnsField->type=TYPE_LOGICAL;
                AnsField->NumData=f2->NumData;
             }       
             else  return ERR_TYPE_MISMATCH;
             break;

       case  OP_THEN:
             if (f1->type==TYPE_LOGICAL) {
                if (f1->NumData || !calc) {  // simply return the second argument 
                   if (f2->type==TYPE_LOGICAL) return ERR_TYPE_MISMATCH;

                   AnsField->DblData=1;      // pass the TRUE value to the ELSE operator
                   AnsField->NumData=1;      // Pass TRUE to the ELSE operator
                   AnsField->type=f2->type;
                   if (f2->type==TYPE_TEXT) {
                      if (calc) {
                         len=lstrlen(f2->CharData);
                         if (NULL==(AnsField->CharData=OurAlloc(len+2))) return ERR_NO_MEM;
                         lstrcpy(AnsField->CharData,f2->CharData);
                      }
                   }
                   else if (f2->type==TYPE_DBL) AnsField->DblData=f2->DblData;
                   else                         AnsField->NumData=f2->NumData;
                }
                else {
                   AnsField->type=f2->type;
                   if (f2->type==TYPE_TEXT) {
                      if (calc) {
                         if (NULL==(AnsField->CharData=OurAlloc(2))) return ERR_NO_MEM;
                         lstrcpy(AnsField->CharData,"");
                      }
                   }
                   AnsField->DblData=0;      // pass the FALSE value to the ELSE operator
                   AnsField->NumData=0;      // Pass FALSE to the ELSE operator
                }
             }       
             else  return ERR_TYPE_MISMATCH;
             break;

       case  OP_ELSE:
             if (!calc && f1->type!=f2->type) return ERR_TYPE_MISMATCH;

             // decide which field to return
             if ( (f1->type!=TYPE_NUM && f1->NumData) 
               || (f1->type==TYPE_NUM && f1->DblData)) fld=f1;  // return the THEN value
             else                                      fld=f2;  // return the ELSE value
             
             AnsField->type=fld->type;
             if (fld->type==TYPE_TEXT) {
                if (calc) {
                   len=lstrlen(fld->CharData);
                   if (NULL==(AnsField->CharData=OurAlloc(len+2))) return ERR_NO_MEM;
                   lstrcpy(AnsField->CharData,fld->CharData);
                }
             }
             else {
                AnsField->NumData=fld->NumData;
                AnsField->DblData=fld->DblData;
             }
             break;

       default:
            return ERR_INCOMPLETE;      // operator not processed 
    }

    return 0;
}

/******************************************************************************
    calculate3:
    This routine performs calculation for all function type operators.
    The first and second argument specify the two input field. The third
    argument is used to return the result.
    This function returns 0 if the operator is processed, otherwise
    it returns an error code.
******************************************************************************/
calculate3(struct StrField huge *f1,struct StrField huge *f2,struct StrField huge *AnsField,int op,int type1,int type2,int calc)
{
    int  len,i,count;
    long mm,dd,yy,days;
    char string[LINE_WIDTH+2];
       
    switch (op) {
       case  OP_LEN:
             if (f2->type==TYPE_TEXT) {
                AnsField->type=TYPE_NUM;
                if (calc) {
                   if (f2->CharData==NULL) AnsField->NumData=0;
                   else                    AnsField->NumData=lstrlen(f2->CharData);
                }
             }       
             else  return ERR_TYPE_MISMATCH;
             break;

       case  OP_UPPER:
       case  OP_LOWER:
       case  OP_TRIM:
             if (f2->type==TYPE_TEXT) {
                AnsField->type=f2->type;
                if (calc) {
                    len=lstrlen(f2->CharData);
                    if (NULL==(AnsField->CharData=OurAlloc(len+2))) return ERR_NO_MEM;
                    lstrcpy(AnsField->CharData,f2->CharData);
                    if      (op==OP_UPPER) lstrupr(AnsField->CharData);
                    else if (op==OP_LOWER) lstrlwr(AnsField->CharData);
                    else                   StrTrim(AnsField->CharData);
                }
             }       
             else  return ERR_TYPE_MISMATCH;
             break;

       case  OP_WORD:
             if (f1->type==TYPE_TEXT) {
                if      (f2->type==TYPE_NUM) count=(int)f2->NumData;
                else if (f2->type==TYPE_DBL) count=(int)f2->DblData;
                else return ERR_TYPE_MISMATCH;

                AnsField->type=TYPE_TEXT;
                if (calc) {
                   if (!ExtractSpecificWord(f1->CharData,string,count)) return ERR_NO_MEM;
                   len=strlen(string);
                   if (NULL==(AnsField->CharData=OurAlloc(len+2))) return ERR_NO_MEM;
                   lstrcpy(AnsField->CharData,string);
                }
             }       
             else  return ERR_TYPE_MISMATCH;
             break;

       case  OP_CHAR:
             if (f1->type==TYPE_TEXT) {
                if      (f2->type==TYPE_NUM) count=(int)f2->NumData;
                else if (f2->type==TYPE_DBL) count=(int)f2->DblData;
                else return ERR_TYPE_MISMATCH;

                AnsField->type=TYPE_TEXT;
                if (calc) {
                    if (NULL==(AnsField->CharData=OurAlloc(2))) return ERR_NO_MEM;
                    len=lstrlen(f1->CharData);
                    if (count<1 || count>len) AnsField->CharData[0]=0;
                    else {
                       AnsField->CharData[0]=f1->CharData[count-1];
                       AnsField->CharData[1]=0;
                    }
                }
             }       
             else  return ERR_TYPE_MISMATCH;
             break;

       case  OP_TEXT:                      // convert to text format
             AnsField->type=TYPE_TEXT;
             if (calc) {
                FormatFieldData(f2,string);
                if (f2->type==TYPE_TEXT) len=lstrlen(f2->CharData);
                else                     len=lstrlen(string);
                if (NULL==(AnsField->CharData=OurAlloc(len+2))) return ERR_NO_MEM;
                if (f2->type==TYPE_TEXT) lstrcpy(AnsField->CharData,f2->CharData);
                else                     lstrcpy(AnsField->CharData,string);
             }
             break;

       case  OP_FIRST_CHARS:               // extract first specified number of characters
             if (f1->type==TYPE_TEXT) {
                if      (f2->type==TYPE_NUM) count=(int)f2->NumData;
                else if (f2->type==TYPE_DBL) count=(int)f2->DblData;
                else return ERR_TYPE_MISMATCH;

                AnsField->type=TYPE_TEXT;
                if (calc) {
                   if (count>lstrlen(f1->CharData)) count=lstrlen(f1->CharData);
                   lstrcpy(string,f1->CharData);
                   string[count]=0;
                   if (NULL==(AnsField->CharData=OurAlloc(count+2))) return ERR_NO_MEM;
                   lstrcpy(AnsField->CharData,string);
                }
             }       
             else  return ERR_TYPE_MISMATCH;
             break;

       case  OP_LAST_CHARS:               // extract last specified number of characters
             if (f1->type==TYPE_TEXT) {
                if      (f2->type==TYPE_NUM) count=(int)f2->NumData;
                else if (f2->type==TYPE_DBL) count=(int)f2->DblData;
                else return ERR_TYPE_MISMATCH;

                AnsField->type=TYPE_TEXT;
                if (calc) {
                   len=lstrlen(f1->CharData);
                   if (count>len) count=len;
                   lstrcpy(string,&(f1->CharData[len-count]));
                   string[count]=0;
                   if (NULL==(AnsField->CharData=OurAlloc(count+2))) return ERR_NO_MEM;
                   lstrcpy(AnsField->CharData,string);
                }
             }       
             else  return ERR_TYPE_MISMATCH;
             break;

       case  OP_MIN:
       case  OP_MAX:
             if (type1!=TYPE_NUM && type1!=TYPE_DBL) return ERR_TYPE_MISMATCH;
             if (type2!=TYPE_NUM && type2!=TYPE_DBL) return ERR_TYPE_MISMATCH;
             if (type1==TYPE_DBL && type2!=TYPE_DBL) f2->DblData=(double)f2->NumData;
             if (type2==TYPE_DBL && type1!=TYPE_DBL) f1->DblData=(double)f1->NumData;
             if (type1==type2) AnsField->type=type1;
             else              AnsField->type=TYPE_DBL;
             if (op==OP_MIN) {
                AnsField->NumData=min(f1->NumData,f2->NumData);
                AnsField->DblData=min(f1->DblData,f2->DblData);
             }
             else {
                AnsField->NumData=max(f1->NumData,f2->NumData);
                AnsField->DblData=max(f1->DblData,f2->DblData);
             }
             break;

       case  OP_ROUND:
             if (type1!=TYPE_NUM && type1!=TYPE_DBL) return ERR_TYPE_MISMATCH;
             if (type2!=TYPE_NUM && type2!=TYPE_DBL) return ERR_TYPE_MISMATCH;
             if (type1!=TYPE_DBL) f1->DblData=(double)f1->NumData;
             if (type2!=TYPE_NUM) f2->NumData=(long)f2->DblData;
             AnsField->type=TYPE_DBL;
             AnsField->DblData=f1->DblData;
             for (i=0;i<f2->NumData;i++) AnsField->DblData=(AnsField->DblData)*10;
             AnsField->DblData=(double)((long)AnsField->DblData);
             for (i=0;i<f2->NumData;i++) AnsField->DblData=(AnsField->DblData)/10;
             break;

       case  OP_INT:
             AnsField->type=TYPE_NUM;
             if (calc) {
                if   (type2==TYPE_NUM || type2==TYPE_LOGICAL || type2==TYPE_DATE) 
                                          AnsField->NumData=f2->NumData;
                else if (type2==TYPE_DBL) AnsField->NumData=(long)f2->DblData;
                else                      AnsField->NumData=atol(f2->CharData);
             }
             break;

       case  OP_ABS:
             if (type2!=TYPE_NUM && type2!=TYPE_DBL) return ERR_TYPE_MISMATCH;
             AnsField->type=type2;
             AnsField->NumData=f2->NumData;
             AnsField->DblData=f2->DblData;
             if (type2==TYPE_NUM && f2->NumData<0) AnsField->NumData=-(f2->NumData);
             if (type2==TYPE_DBL && f2->DblData<0) AnsField->DblData=-(f2->DblData);
             break;
       
       case  OP_WEEKDAY:
             if (type2!=TYPE_DATE) return ERR_TYPE_MISMATCH;
             AnsField->type=TYPE_TEXT;
             yy=(f2->NumData)/10000L;
             mm=(f2->NumData-yy*10000L)/100L;
             dd=f2->NumData-yy*10000L-mm*100L;
             
             yy=yy-1900;
             days=yy*365;                   // calculate days since 12/31/1899 
             if (yy>0) days=days+((yy-1)/4)+1; // add leap days 
             for (i=1;i<mm;i++) {           // add days for the previous months 
                if (i==2) {                 // feb 
                   if (((yy/4)*4)==yy) days=days+29;
                   else                days=days+28;
                }
                else if (i==4 || i==6 || i==9 || i==11) days=days+30;
                else                                    days=days+31;
             }
             days=days+dd;                  // add the day of the month 

             days=days+6;                   // adjustment factor 

             days=days-((days/7)*7);        // week day 

             if (calc) {                    // store the week day 
                if (NULL==(AnsField->CharData=OurAlloc(12))) return ERR_NO_MEM;
                if (days==0) lstrcpy(AnsField->CharData,"Sunday");
                if (days==1) lstrcpy(AnsField->CharData,"Monday");
                if (days==2) lstrcpy(AnsField->CharData,"Tuesday");
                if (days==3) lstrcpy(AnsField->CharData,"Wednesday");
                if (days==4) lstrcpy(AnsField->CharData,"Thursday");
                if (days==5) lstrcpy(AnsField->CharData,"Friday");
                if (days==6) lstrcpy(AnsField->CharData,"Saturday");
             }
             break;

       case  OP_DAY:
             if (type2!=TYPE_DATE) return ERR_TYPE_MISMATCH;
             AnsField->type=TYPE_NUM;
             yy=(f2->NumData)/10000L;
             mm=(f2->NumData-yy*10000L)/100L;
             AnsField->NumData=f2->NumData-yy*10000L-mm*100L;
             break;

       case  OP_MONTH:
             if (type2!=TYPE_DATE) return ERR_TYPE_MISMATCH;
             AnsField->type=TYPE_NUM;
             yy=(f2->NumData)/10000L;
             AnsField->NumData=(f2->NumData-yy*10000L)/100L;
             break;

       case  OP_YEAR:
             if (type2!=TYPE_DATE) return ERR_TYPE_MISMATCH;
             AnsField->type=TYPE_NUM;
             AnsField->NumData=(f2->NumData)/10000L;
             break;

       default:
             return ERR_INCOMPLETE;              // operator not processed 
    }

    return 0;
}

/******************************************************************************
    ExtractSpecificWord:
    Extract a word form a word string.
******************************************************************************/
ExtractSpecificWord(LPSTR CurString,LPSTR CurWord,int WordNum)
{
    int   len;
    LPSTR string,tok,PrevTok=NULL;
    
    CurWord[0]=0;
    if (WordNum<1) return TRUE;

    if (CurString==NULL) len=0;
    else                 len=lstrlen(CurString);
    
    if (NULL==(string=OurAlloc(len+2))) return FALSE;
    if (CurString) lstrcpy(string,CurString);
    else           lstrcpy(string,"");

    if (NULL==(tok=_fstrtok(string," ,;."))) goto END;
    WordNum--;

    while(TRUE) {
        if (WordNum==0) goto END;                   // word located 
        PrevTok=tok;
        if (NULL==(tok=_fstrtok(NULL," ,;."))) break; // end of string 
        WordNum--;
    }

    if (PrevTok!=NULL) tok=PrevTok;                 // return the last word 

    END:
    if (tok!=NULL) {
       if (lstrlen(tok)>NAME_WIDTH) tok[NAME_WIDTH]=0;
       lstrcpy(CurWord,tok);
    }
    OurFree(string);
    return TRUE;
}

